%% function [best_paths, best_paths_weights, best_paths_index, n_paths, all_paths,...
%    mW, sigmaW, rhoW, dW] = find_best_paths(H, R, clutter_rate,...
%    xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, M_max, Z)
% Function to find best observation subsets (paths) for each Gaussian component in a greedy manner
% The input parameters are:
% H: observation matrix
% R: observation noise covariance
% clutter_rate: poisson mean clutter rate
% xrange, yrange: monitoring region dimensions in meters
% p_detect: a vector of probability of detection of the sensors (assumed constant
%    over the complete state space); length of this vector is equal to the number of sensors
% The predicted PHD is modeled as a Gaussian mixture model with parameters w_k1, m_k1, P_k1, J_k1
% J_k1: number of Gaussian components in predicted PHD
% w_k1, m_k1, P_k1: weights, means and covariance matrices of the Gaussian components in predicted PHD
% M_max: 
% Z: set of observations generated by the sensor
% The function outputs ...
% best_paths: the collection of best observation subsets for each of the J_k1 Gaussian components 
% best_paths_weights: the path weight dW without the clutter term 
% all_paths: collection of all unique paths for all the components
% n_paths: number of paths in all_paths
% best_paths_index: the index of paths for each component in the collection all_paths 
% mW, sigmaW: means and covariance matrices of Gaussian components in rho_W x r_{k+1|k}
% rhoW: the weight of Gaussian components in rho_W x s_{k+1|k}
% dW: the weight associated with the observation path W


function [best_paths, best_paths_weights, best_paths_index, n_paths, all_paths,...
    mW, sigmaW, rhoW, dW] = find_best_paths(H, R, clutter_rate,...
    xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, M_max, Z)

clutter_density = 1/((xrange(2)-xrange(1))*(yrange(2)-yrange(1)));  % uniform clutter density
no_sensors = size(Z,2);

% initialize output variables
best_paths = cell(1,J_k1);
best_paths_weights = cell(1,J_k1);

% for each Gaussian component
for j = 1:J_k1
    curr_J = 1;
    curr_w = w_k1(j);
    curr_m = m_k1(:,j);
    curr_P = P_k1(:,:,j);
    curr_paths = [];
    
    % process each sensor sequentially
    for s = 1:no_sensors
        curr_Z = Z{s};
        no_meas = size(curr_Z,2);    % current number of measurements

        curr_S = zeros(size(R));
        curr_K = zeros(size(H'));
        curr_P_k_k = zeros(size(curr_P));
        
        curr_eta = H*curr_m;
        for i = 1:curr_J
            curr_S(:,:,i) = R + H*curr_P(:,:,i)*H';
            inverseS = curr_S(:,:,i) \ eye(2);
            curr_K(:,:,i) = curr_P(:,:,i)*H'*inverseS;
            curr_P_k_k(:,:,i) = (eye(4,4)-curr_K(:,:,i)*H)*curr_P(:,:,i);
        end
    
        % calculate q(z,i) seperately for efficiency
        q = zeros(no_meas,curr_J);
        for i = 1:curr_J
            q(:,i) = mvnpdf(curr_Z',curr_eta(:,i)',curr_S(:,:,i));
        end

        % for no detected observation case
        local_w = curr_w*(1-p_detect(s));
        local_m = curr_m;
        local_P = curr_P;
        local_paths = [curr_paths -1*ones(curr_J,1)];
        
        % for detected observations
        for z = 1:no_meas
            index = z*curr_J + (1:curr_J);
            local_w(index) = p_detect(s)*curr_w.*q(z,:)/clutter_density;
            local_P(:,:,index) = curr_P_k_k;
            for i = 1:curr_J
                local_m(:,index(i)) = curr_m(:,i) + curr_K(:,:,i)*(curr_Z(:,z)-curr_eta(:,i));
            end
            local_paths = [local_paths; [curr_paths z*ones(curr_J,1)]];
        end
        
        % select best observation sequences till now 
        % always include the empty observation path first
        [~,index] = sort(local_w(2:end),'descend');
        J = min(length(index),M_max);  % update number of components propagated
        index = index(1:J) + 1;
        curr_w = [local_w(1) local_w(index)];
        curr_m = [local_m(:,1) local_m(:,index)];
        curr_P = cat(3,local_P(:,:,1),local_P(:,:,index));
        curr_paths = [local_paths(1,:); local_paths(index,:)];
        curr_J = J + 1;
    end
    
    % final paths information for j_th Gaussian component
    best_paths{j} = curr_paths;
end


%% generate the collection of paths - all_paths
all_paths = [];
for j = 1:J_k1
    [all_paths ] = [all_paths; best_paths{j}];
end

all_paths = unique(all_paths,'rows');
index = sum(all_paths ~= -1,2) == 0;
all_paths(index,:) = [];
n_paths = size(all_paths,1);

best_paths_index = cell(1,J_k1);
for r = 1:J_k1
    for l = 1:size(best_paths{r},1)
        for i = 1:n_paths
            if all_paths(i,:) == best_paths{r}(l,:);
                best_paths_index{r}(l) = i;
                break;
            end
        end
    end
end

%% calculate path weights (rhoW and dW)

% initialize
rhoW = repmat(w_k1,1,n_paths);
mW = repmat(m_k1,1,n_paths);
sigmaW = repmat(P_k1,[1,1,n_paths]);
pathW = zeros(n_paths,1);
dW = zeros(n_paths,1);

for i = 1:n_paths
    temp_w1 = w_k1;  % temp variable for calculating path weights
    temp_w2 = w_k1;  % temp variable for calculating dW (with clutter_rate)
    
    for s = 1:no_sensors
        % vectorize over all components J_k1
        index = (i-1)*J_k1 + (1:J_k1);
        if all_paths(i,s) == -1
            rhoW(index) = rhoW(index)*(1-p_detect(s));
            temp_w1 = temp_w1*(1-p_detect(s));
            temp_w2 = temp_w2*(1-p_detect(s));
        else
            % intermediate parameters
            z = Z{s}(:,all_paths(i,s));
            eta = H*mW(:,index);
            for j = 1:J_k1
                symmetricS(:,:,j) = R + H*sigmaW(:,:,index(j))*H';
                inverseS = symmetricS(:,:,j) \ eye(2);
                K(:,:,j) = sigmaW(:,:,index(j))*H'*inverseS;
                K_eta(:,j) = K(:,:,j)*(z - eta(:,j));
                P(:,:,j) = (eye(4,4)-K(:,:,j)*H)*sigmaW(:,:,index(j));
            end
            q = mvnpdf(z',eta',symmetricS);

            % update parameters (for rhoW and dW)
            rhoW(index) = p_detect(s)*rhoW(index).*q';
            temp_w1 = p_detect(s)*temp_w1.*q'/clutter_density;
            temp_w2 = p_detect(s)*temp_w2.*q'/(clutter_density*clutter_rate);
            mW(:,index) = mW(:,index) + K_eta;
            sigmaW(:,:,index) = P;
        end
    end

    % normalization
    normalization_factor = sum(rhoW(index));
    rhoW(index) = rhoW(index)/normalization_factor;    % rho_W x r_{k+1|k}
    pathW(i) = sum(temp_w1);    % dW (without clutter rate)
    dW(i) = sum(temp_w2);    % dW (with clutter rate)
end


for r = 1:J_k1
    best_paths_weights{r} = ones(1,size(best_paths{r},1));
    for l = 2:size(best_paths{r},1) % first path has weight 1 by convention
        best_paths_weights{r}(l) = pathW(best_paths_index{r}(l));
    end
end


end
