%% function [pmf_k, rhoW, m_k, P_k, J_k] = gen_gm_ms_cphd_update(H, R, clutter_rate,...
%    xrange, yrange, p_detect, pmf_k1, w_k1, m_k1, P_k1, J_k1, W_max, PT_max, Z)
% Gaussian mixture implementation of multi sensor CPHD filter update for linear-Gaussian observation model
% Input parameters are as follows:
% H: observation matrix
% R: observation noise covariance
% clutter_rate: poisson mean clutter rate
% xrange, yrange: monitoring region dimensions in meters
% p_detect: a vector of probability of detection of the sensors (assumed constant
%    over the complete state space); length of this vector is equal to the number of sensors
% pmf_k1: cardinality mass function of the predicted multi-target state
% The predicted PHD is modeled as a Gaussian mixture model with parameters w_k1, m_k1, P_k1, J_k1
% J_k1: number of Gaussian components in predicted PHD
% w_k1, m_k1, P_k1: weights, means and covariance matrices of the Gaussian components in predicted PHD
% W_max: maximum number of observation subsets for the greedy algorithm
% PT_max: maximum number of partitions for the greedy algorithm
% Z: set of observations generated by the sensor
% The function outputs Gaussian mixture component parameters of the posterior PHD
% J_k: number of Gaussian components for posterior PHD
% rhoW, m_k, P_k: weights, means and covariance matrices of the Gaussian components in posterior PHD
% pmf_k: cardinality mass function of the posterior multi-target state

function [pmf_k, rhoW, m_k, P_k, J_k] = gen_gm_ms_cphd_update(H, R, clutter_rate,...
    xrange, yrange, p_detect, pmf_k1, w_k1, m_k1, P_k1, J_k1, W_max, PT_max, Z)

w_k1 = w_k1/sum(w_k1);  % normalize to represent s(x)

num_sensors = size(Z,2);
empty_data = 1;

for s = 1:num_sensors
    if ~isempty(Z{s})
        empty_data = 0;
        break;
    end
end

%% if empty measurement set
if empty_data
    
    gamma = prod(1-p_detect);
    max_card = length(pmf_k1) - 1;
    
    pgf_0 = gamma.^(0:max_card)*pmf_k1';
    pgf_1 = ((1:max_card).*gamma.^(0:max_card-1))*pmf_k1(2:end)';
    
    ratio = (pgf_1/pgf_0)*gamma;
    rhoW = w_k1*ratio;
    m_k = m_k1;
    P_k = P_k1;
    J_k = J_k1;
    pmf_k = pmf_k1;
    
else

    %% Find best observation subsets (paths) for each component
    [trellis_paths, trellis_weights, trellis_path_index, n_paths, ~,...
        m_k, P_k, rhoW, dW] = find_best_paths(H, R, clutter_rate,...
        xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, W_max, Z);
    
    %% Find likely partitions
    [~,partitions_v] = find_partitions(J_k1,trellis_paths,...
        trellis_path_index,trellis_weights,n_paths,PT_max);


    %% General multi-sensor CPHD filter over reduced partitions

    % calculate path weights by summing over corresponding partitions
    % predicted pgf derivatives evaluated at gamma and calculation of psi_P
    gamma = prod(1-p_detect);
    max_card = length(pmf_k1) - 1;
    factorial_table = factorial(0:max_card);
    psi(1) = gamma.^(0:max_card)*pmf_k1';
    for i = 1:n_paths+1
        index = i:max_card;
        nPi = factorial_table(index+1)./factorial_table(index - i + 1);
        gamma_powers = gamma.^(index - i);
        pgf_coeff = nPi.*gamma_powers;
        psi(i+1) = pgf_coeff*pmf_k1(index+1)';   % predicted pgf derivative
    end


    % efficient calculation of path weights
    log_dW = log(dW' + realmin);    % fix to avoid -Inf terms in LOG
    partition_size = sum(partitions_v,2);
    temp = bsxfun(@times,partitions_v,log_dW);
    partition_dw = exp(sum(temp,2));
    partition_w = partition_dw.*psi(partition_size+1)';
    partition_w_star = partition_dw.*psi(partition_size+2)';

    temp = bsxfun(@times,partitions_v,partition_w);
    path_w = sum(temp,1)';
    path_w = path_w/sum(partition_w);

    for i = 1:n_paths
        for j = 1:J_k1
            rhoW((i-1)*J_k1+j) = rhoW((i-1)*J_k1+j)*path_w(i);
        end
    end

    % case of no target detection by any sensor (PHD update)
    ratio = (sum(partition_w_star)/sum(partition_w))*prod(1-p_detect);
    rhoW = [rhoW w_k1*ratio];
    m_k = [m_k m_k1];
    P_k = cat(3,P_k,P_k1);
    J_k = J_k1*(1 + n_paths);

    %% posterior cardinality distribution update
    beta = zeros(1,max_card+1);
    pmf_k = zeros(1,max_card+1);
    partition_dw_size = zeros(1,n_paths+1);
    gamma_powers = gamma.^(0:max_card);
    for j = 0:n_paths
        index = partition_size == j;
        partition_dw_size(j+1) = sum(partition_dw(index));
    end

    for n = 0:max_card
        for j = 0:min(n,n_paths)
            beta(n+1) = beta(n+1) + partition_dw_size(j+1)*gamma_powers(n-j+1)*factorial_table(n+1)/factorial_table(n-j+1);
        end
        pmf_k(n+1) = beta(n+1)*pmf_k1(n+1);
    end
    % normalization of pmf
    pmf_k = pmf_k/sum(pmf_k);
end

end

