%% function [rhoW, m_k, P_k, J_k] = gen_gm_ms_phd_update(H, R, clutter_rate,...
%    xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, W_max, PT_max, Z)
% Gaussian mixture implementation of multi sensor PHD filter update for linear-Gaussian observation model
% Input parameters are as follows:
% H: observation matrix
% R: observation noise covariance
% clutter_rate: poisson mean clutter rate
% xrange, yrange: monitoring region dimensions in meters
% p_detect: a vector of probability of detection of the sensors (assumed constant
%    over the complete state space); length of this vector is equal to the number of sensors
% The predicted PHD is modeled as a Gaussian mixture model with parameters w_k1, m_k1, P_k1, J_k1
% J_k1: number of Gaussian components in predicted PHD
% w_k1, m_k1, P_k1: weights, means and covariance matrices of the Gaussian components in predicted PHD
% W_max: maximum number of observation subsets for the greedy algorithm
% PT_max: maximum number of partitions for the greedy algorithm
% Z: set of observations generated by the sensor
% The function outputs Gaussian mixture component parameters of the posterior PHD
% J_k: number of Gaussian components for posterior PHD
% rhoW, m_k, P_k: weights, means and covariance matrices of the Gaussian components in posterior PHD

function [rhoW, m_k, P_k, J_k] = gen_gm_ms_phd_update(H, R, clutter_rate,...
    xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, W_max, PT_max, Z)

num_sensors = size(Z,2);
empty_data = 1;

for s = 1:num_sensors
    if ~isempty(Z{s})
        empty_data = 0;
        break;
    end
end

%% if empty measurement set
if empty_data
    rhoW = w_k1*prod(1-p_detect);
    m_k = m_k1;
    P_k = P_k1;
    J_k = J_k1;
    
else
    
    %% Find best observation subsets (paths) for each component
    [trellis_paths, trellis_weights, trellis_path_index, n_paths, ~,...
        m_k, P_k, rhoW, dW] = find_best_paths(H, R, clutter_rate,...
        xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, W_max, Z);
    
    %% Find likely partitions
    [~,partitions_v] = find_partitions(J_k1,trellis_paths,...
        trellis_path_index,trellis_weights,n_paths,PT_max);


    %% General multi-sensor PHD filter over reduced partitions

    % efficient calculation of path weights
    log_dW = log(dW' + realmin);    % fix to avoid -Inf terms in LOG
    temp = bsxfun(@times,partitions_v,log_dW);
    partition_w = exp(sum(temp,2));

    temp = bsxfun(@times,partitions_v,partition_w);
    path_w = sum(temp,1)';
    path_w = path_w/sum(partition_w);

    for i = 1:n_paths
        for j = 1:J_k1
            rhoW((i-1)*J_k1+j) = rhoW((i-1)*J_k1+j)*path_w(i);
        end
    end

    %% case of no target detection by any sensor
    rhoW = [rhoW w_k1*prod(1-p_detect)];
    m_k = [m_k m_k1];
    P_k = cat(3,P_k,P_k1);
    J_k = J_k1*(1 + n_paths);
end

end

