% function [pmf_k, w_k, m_k, P_k, J_k] = gm_cphd_update(H, R, clutter_rate,...
%    xrange, yrange, p_detect, pmf_k1, w_k1, m_k1, P_k1, J_k1, Z)
% Gaussian mixture implementation of single sensor CPHD update for linear-Gaussian observation model
% Input parameters are as follows:
% H: observation matrix
% R: observation noise covariance
% clutter_rate: poisson mean clutter rate
% xrange, yrange: monitoring region dimensions in meters
% p_detect: single sensor probability of detection
% pmf_k1: cardinality mass function of the predicted multi-target state
% The predicted PHD is modeled as a Gaussian mixture model with parameters w_k1, m_k1, P_k1, J_k1
% J_k1: number of Gaussian components in predicted PHD
% w_k1, m_k1, P_k1: weights, means and covariance matrices of the Gaussian components in predicted PHD
% Z: set of observations generated by the sensor
% The function outputs Gaussian mixture component parameters of the posterior PHD
% J_k: number of Gaussian components for posterior PHD
% w_k, m_k, P_k: weights, means and covariance matrices of the Gaussian components in posterior PHD
% pmf_k: cardinality mass function of the posterior multi-target state

function [pmf_k, w_k, m_k, P_k, J_k] = gm_cphd_update(H, R, clutter_rate,...
    xrange, yrange, p_detect, pmf_k1, w_k1, m_k1, P_k1, J_k1, Z)

if isempty(Z)   % if empty measurement set
    w_k = (1-p_detect)*w_k1;
    m_k = m_k1;
    P_k = P_k1;
    J_k = J_k1;
    pmf_k = pmf_k1;

else
    % Calculate required statistics
    % normalize weights
    w_k1 = w_k1/sum(w_k1);

    clutter_density = 1/((xrange(2)-xrange(1))*(yrange(2)-yrange(1)));  % uniform clutter density
    
    % means and variances of components    
    symmetricS = zeros(size(R));
    K_k = zeros(size(H'));
    P_k_k = zeros(size(P_k1));

    eta_k1 = H*m_k1;
    for j = 1:J_k1
        symmetricS(:,:,j) = R + H*P_k1(:,:,j)*H';
        inverseS = symmetricS(:,:,j) \ eye(2);
        K_k(:,:,j) = P_k1(:,:,j)*H'*inverseS;
        P_k_k(:,:,j) = (eye(4,4)-K_k(:,:,j)*H)*P_k1(:,:,j);
    end

    % number of measurements
    num_meas = size(Z,2);

    % clutter pgf derivatives evaluated at zero
    index = 0:num_meas;
    pmf_c =  poisspdf(index,clutter_rate);
    pgf_c = factorial(index).*pmf_c(index+1);
    pgf_c_lr = fliplr(pgf_c);

    % gamma parameter
    gamma = 1 - p_detect;

    % predicted pgf derivatives evaluated at gamma
    max_card = length(pmf_k1) - 1;
    pgf_k1(1) = gamma.^(0:max_card)*pmf_k1';
    for i = 1:num_meas + 1
        index = i:max_card;
        nPi = factorial(index)./factorial(index - i);
        gamma_powers = gamma.^(index - i);
        pgf_coeff = nPi.*gamma_powers;
        pgf_k1(i+1) = pgf_coeff*pmf_k1(index+1)';   % predicted pgf derivative
    end

    % calculate q(z,j) seperately for efficiency
    q = zeros(num_meas,J_k1);
    for j = 1:J_k1
        q(:,j) = mvnpdf(Z',eta_k1(:,j)',symmetricS(:,:,j));
    end

    % symmetric functions
    Y = zeros(1,num_meas);
    for z = 1:num_meas
        temp = p_detect*w_k1*q(z,:)';
        Y(z) = temp/clutter_density;
    end

    % relation between elementary symmetric function and polynomials
    % abs() used because Y has all positive entries
    sigma_i = abs(poly(Y));
    for i = 1:num_meas
        Y_temp = Y;
        Y_temp(i) = [];
        sigma_z(i,:) = abs(poly(Y_temp));
    end

    % alpha parameters (for PHD update)
    temp_nume = pgf_c_lr.*pgf_k1(2:num_meas+2);
    temp_nume = temp_nume*sigma_i';

    temp_deno = pgf_c_lr.*pgf_k1(1:num_meas+1);
    temp_deno = temp_deno*sigma_i';

    alpha_0 = temp_nume/temp_deno;

    for i = 1:num_meas
        temp_nume = pgf_c_lr(2:num_meas+1).*pgf_k1(2:num_meas+1);
        temp_nume = temp_nume*sigma_z(i,:)';

        temp_deno = pgf_c_lr.*pgf_k1(1:num_meas+1);
        temp_deno = temp_deno*sigma_i';

        alpha_z(i) = temp_nume/temp_deno;
    end

    % beta parameters (for cardinality update)
    for n = 0:max_card
        index = 0:min(n,num_meas);
        nPi = factorial(n)./factorial(n - index);
        gamma_powers = gamma.^(n - index);
        pgf_coeff = nPi.*gamma_powers;

        temp = pgf_c_lr(index+1).*pgf_coeff;
        beta(n+1) = temp*sigma_i(index+1)';
    end

    %% update PHD 
    % Components associated with no detection
    m_k = m_k1;
    P_k = P_k1;
    w_k = (1-p_detect)*alpha_0*w_k1;

    % Components associated with detections
    for z = 1:num_meas
        index = z*J_k1 + (1:J_k1);
        P_k(:,:,index) = P_k_k;
        w_k(index) = p_detect*alpha_z(z)*w_k1.*q(z,:)/clutter_density;
        
        for j = 1:J_k1
            m_k(:,z*J_k1+j) = m_k1(:,j) + K_k(:,:,j)*(Z(:,z)-eta_k1(:,j));
        end
    end

    % Final number of Gaussian components
    J_k = (1+num_meas)*J_k1;

    %% update cardinality
    pmf_k = beta.*pmf_k1;
    pmf_k = pmf_k/sum(pmf_k);
end

end
