%% function [w_k, m_k, P_k, J_k] = gm_phd_update(H, R, clutter_rate,...
%    xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, Z)
% Gaussian mixture implementation of single sensor PHD filter update for linear-Gaussian observation model
% Input parameters are as follows:
% H: observation matrix
% R: observation noise covariance
% clutter_rate: poisson mean clutter rate
% xrange, yrange: monitoring region dimensions in meters
% p_detect: single sensor probability of detection
% The predicted PHD is modeled as a Gaussian mixture model with parameters w_k1, m_k1, P_k1, J_k1
% J_k1: number of Gaussian components in predicted PHD
% w_k1, m_k1, P_k1: weights, means and covariance matrices of the Gaussian components in predicted PHD
% Z: set of observations generated by the sensor
% The function outputs Gaussian mixture component parameters of the posterior PHD
% J_k: number of Gaussian components for posterior PHD
% w_k, m_k, P_k: weights, means and covariance matrices of the Gaussian components in posterior PHD

function [w_k, m_k, P_k, J_k] = gm_phd_update(H, R, clutter_rate,...
    xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, Z)

if isempty(Z)   % if empty measurement set
    w_k = (1-p_detect)*w_k1;
    m_k = m_k1;
    P_k = P_k1;
    J_k = J_k1;
else
    clutter_density = 1/((xrange(2)-xrange(1))*(yrange(2)-yrange(1)));  % uniform clutter density
    
    symmetricS = zeros(size(R));
    K_k = zeros(size(H'));
    P_k_k = zeros(size(P_k1));

    eta_k1 = H*m_k1;
    for j = 1:J_k1
        eta_k1(:,j) = H*m_k1(:,j);
        symmetricS(:,:,j) = R + H*P_k1(:,:,j)*H';
        inverseS =  symmetricS(:,:,j) \ eye(2);
        K_k(:,:,j) = P_k1(:,:,j)*H'*inverseS;
        P_k_k(:,:,j) = (eye(4,4)-K_k(:,:,j)*H)*P_k1(:,:,j);
    end

    % Components associated with no detection
    w_k = (1-p_detect)*w_k1;
    m_k = m_k1;
    P_k = P_k1;

    % calculate q(z,j) seperately for efficiency
    num_meas = size(Z,2);
    q = zeros(num_meas,J_k1);
    for j = 1:J_k1
        q(:,j) = mvnpdf(Z',eta_k1(:,j)',symmetricS(:,:,j));
    end
    
    kappa = clutter_rate*clutter_density;
    % Components associated with detections
    for z = 1:num_meas
        index = z*J_k1 + (1:J_k1);
        w_k(index) = p_detect*w_k1.*q(z,:);
        P_k(:,:,index) = P_k_k;
        
        for j = 1:J_k1
            m_k(:,index(j)) = m_k1(:,j) + K_k(:,:,j)*(Z(:,z)-eta_k1(:,j));
        end
        
        sum_w = sum(w_k(index));
        w_k(index) = w_k(index)/(kappa + sum_w);
    end

    J_k = (1+num_meas)*J_k1;
end

end
