% output_s = mainfun(algorithm,pD,do_reverse,W_max,PT_max,clutter_rate,seed)
% Function to generate target trajectories, sensor observations and perform tracking. 
% The input variables are as follows:
% algorithm: integer taking values in the range {1,2,3,4} to select the
%   tracking algorithms as indicated below 
%   1 multisensor IC PHD filter
%   2 general multisensor PHD filter
%   3 multisensor IC CPHD filter
%   4 general multisensor CPHD filter
% pD: a vector of probability of detection of the sensors (assumed constant
%   over the complete state space); length of this vector is equal to the number of sensors
% do_reverse: indicates order in which sensors are processed; if zero, pD is
%   processed from left to right; if one, pD is processed from right to left
% W_max: maximum number of observation subsets for the greedy algorithms
% PT_max: maximum number of partitions for the greedy algorithms
% clutter_rate: clutter is assumed to be Poisson with uniform spatial
%   density and mean cardinality of clutter_rate
% seed: initialization seed for the random number generator
% output_s: output structure which contains information about true and
%   estimated target states

function output_s = mainfun(algorithm,pD,do_reverse,W_max,PT_max,clutter_rate,seed)

try
    tic;
    %% initialize parameters
    T = 1; % sampling period
    A = [eye(2,2) T*eye(2,2); zeros(2,2) eye(2,2)]; % state transition matrix
    B = [(T^2/2)*eye(2,2); T*eye(2,2)]; % process noise matrix

    p_survive = 0.99; % survival probability for each target

    H = [eye(2,2) zeros(2,2)]; % observation matrix
    sigma_e = 10; % measurement noise std
    R = sigma_e^2*eye(2,2); % observation noise covariance

    J_max = 100; % max no of Gaussian mixture components
    w_threshold = 10^-5;
    merge_threshold = 4;

    max_card = 20; % max cardinality for cphd filter

    % monitoring region dimensions in meters
    xrange = [-1000 1000];
    yrange = [-1000 1000];

    num_sensors = length(pD); % number of sensors

    
    %% true target dynamics
    % generate target trajectories
    % The _gamma parameters correspond to the target birth model. 
    % The target birth PHD is assumed to be a mixture of Gaussians with following parameters:
    % J_gamma - number of targets/Gaussian components in birth density
    % w_gamma - weights of the Gaussian components
    % m_gamma - means of the Gaussian components
    % P_gamma - covariance matrix (same for all the components)
    
    motion_model = 2;

    if motion_model == 1
        track_seed = 100;
        track_stream = RandStream('mt19937ar','seed',track_seed);

        sigma_v = 0.75; % process noise std
        Q = [(T^4/4)*eye(2,2) (T^3/2)*eye(2,2); (T^3/2)*eye(2,2) (T^2)*eye(2,2)];
        Q = sigma_v^2 * Q; % process noise covariance
    
        xy_sym = 400;

        % birth model
        J_gamma = 4; % number of births
        w_gamma = [0.1 0.1 0.1 0.1];
        m_gamma = [xy_sym xy_sym 0 0; -xy_sym -xy_sym 0 0;...
            xy_sym -xy_sym 0 0; -xy_sym xy_sym 0 0]';
        P_gamma = diag([100 100 25 25]);

        time_slot = 20;
        k_max = 5*time_slot + 1; % no of iterations
        N_max = 8;
        num_targets = zeros(1,k_max);
        target_state = zeros(4,N_max,k_max);
        alive_matrix = zeros(N_max,k_max);

        motion_noise_all = randn(track_stream,2,N_max,k_max);

        target_state(:,1:2,1) = [xy_sym xy_sym -2 -5; -xy_sym -xy_sym 2 -3]';
        num_targets(1) = 2;
        alive_matrix(1:num_targets(1),1) = 1;
        start_index = [1 1];

        for k = 2:k_max
            % propagate targets
            num_targets(k) = num_targets(k-1);
            for i = 1:num_targets(k)
                motion_noise = sigma_v*motion_noise_all(:,i,k - start_index(i));
                target_state(:,i,k) = A*target_state(:,i,k-1) + B*motion_noise;
            end

            % add new targets
            if k == time_slot + 2
                target_state(:,3:4,k) = [-xy_sym xy_sym -5 -7; xy_sym -xy_sym 5 -3]';
                num_targets(k) = 4;
                start_index(3:4) = k;
            elseif k == 2*time_slot + 2
                target_state(:,5:6,k) = [xy_sym xy_sym -3 -4; -xy_sym -xy_sym -4 12]';
                num_targets(k) = 6;
                start_index(5:6) = k;
            elseif k == 3*time_slot + 2
                target_state(:,7:8,k) = [-xy_sym xy_sym 10 3; xy_sym -xy_sym -3 10]';
                num_targets(k) = 8;
                start_index(7:8) = k;
            elseif k == 4*time_slot + 2
                target_state(:,7:8,k) = zeros(4,2);
                num_targets(k) = 6;
            end
            alive_matrix(1:num_targets(k),k) = 1;
        end
        
    elseif motion_model == 2
        % discretized nearly constant velocity model (8 targets)
        track_seed = 101;
        stream = RandStream('mt19937ar','seed',track_seed);
        RandStream.setGlobalStream(stream);
        
        % process noise parameters
        sigma_v = 0.25; % process noise std
        Q = [(T^3/3)*eye(2,2) (T^2/2)*eye(2,2); (T^2/2)*eye(2,2) (T)*eye(2,2)];
        Q = sigma_v^2 * Q; % process noise covariance

        xy_sym = 400;

        % birth model
        J_gamma = 4; % number of births
        w_gamma = [0.1 0.1 0.1 0.1];
        m_gamma = [xy_sym xy_sym 0 0; -xy_sym -xy_sym 0 0; xy_sym -xy_sym 0 0; -xy_sym xy_sym 0 0]';
        P_gamma = diag([100 100 25 25]);

        time_slot = 20;
        k_max = 5*time_slot + 1; % no of iterations
        N_max = 8;
        num_targets = zeros(1,k_max);
        target_state = zeros(4,N_max,k_max);
        alive_matrix = zeros(N_max,k_max);

        target_state(:,1:2,1) = [xy_sym xy_sym 0 -5; -xy_sym -xy_sym -3 -4]';
        num_targets(1) = 2;
        alive_matrix(1:num_targets(1),1) = 1;
        start_index = [1 1];

        for k = 2:k_max
            % propagate targets
            num_targets(k) = num_targets(k-1);
            motion_noise = mvnrnd([0,0,0,0],Q,num_targets(k));

            target_state(:,1:num_targets(k),k) = A*target_state(:,1:num_targets(k),k-1);
            target_state(:,1:num_targets(k),k) = target_state(:,1:num_targets(k),k) + motion_noise';

            % add new targets
            if k == time_slot + 2
                target_state(:,3:4,k) = [-xy_sym xy_sym -4 -8; xy_sym -xy_sym 5 0]';
                num_targets(k) = 4;
                start_index(3:4) = k;
            elseif k == 2*time_slot + 2
                target_state(:,5:6,k) = [xy_sym xy_sym -4 -5; -xy_sym -xy_sym -4 12]';
                num_targets(k) = 6;
                start_index(5:6) = k;
            elseif k == 3*time_slot + 2
                target_state(:,7:8,k) = [-xy_sym xy_sym 10 3; xy_sym -xy_sym -3 10]';
                num_targets(k) = 8;
                start_index(7:8) = k;
            elseif k == 4*time_slot + 2
                target_state(:,7:8,k) = zeros(4,2);
                num_targets(k) = 6;
            end
            alive_matrix(1:num_targets(k),k) = 1;
        end
        
    elseif motion_model == 3
        % discretized nearly constant velocity model (4 targets)
        track_seed = 101;
        stream = RandStream('mt19937ar','seed',track_seed);
        RandStream.setGlobalStream(stream);

        % process noise parameters
        sigma_v = 0.25; % process noise std
        Q = [(T^3/3)*eye(2,2) (T^2/2)*eye(2,2); (T^2/2)*eye(2,2) (T)*eye(2,2)];
        Q = sigma_v^2 * Q; % process noise covariance

        xy_sym = 400;

        % birth model
        J_gamma = 2; % number of births
        w_gamma = [0.1 0.1];
        m_gamma = [xy_sym xy_sym 0 0; -xy_sym -xy_sym 0 0]';
        P_gamma = diag([100 100 25 25]);

        time_slot = 20;
        k_max = 5*time_slot + 1; % no of iterations
        N_max = 4;
        num_targets = zeros(1,k_max);
        target_state = zeros(4,N_max,k_max);
        alive_matrix = zeros(N_max,k_max);

        target_state(:,1,1) = [xy_sym xy_sym -7 1]';
        num_targets(1) = 1;
        alive_matrix(1:num_targets(1),1) = 1;

        for k = 2:k_max
            % propagate targets
            num_targets(k) = num_targets(k-1);
            motion_noise = mvnrnd([0,0,0,0],Q,num_targets(k));

            target_state(:,1:num_targets(k),k) = A*target_state(:,1:num_targets(k),k-1);
            target_state(:,1:num_targets(k),k) = target_state(:,1:num_targets(k),k) + motion_noise';

            % add new targets
            if k == time_slot + 2
                target_state(:,2,k) = [-xy_sym -xy_sym 4 6]';
                num_targets(k) = 2;
            elseif k == 2*time_slot + 2
                target_state(:,3,k) = [xy_sym xy_sym 1 -7]';
                num_targets(k) = 3;
            elseif k == 3*time_slot + 2
                target_state(:,4,k) = [-xy_sym -xy_sym 4 -5]';
                num_targets(k) = 4;
            elseif k == 4*time_slot + 2
                target_state(:,4,k) = zeros(4,1);
                num_targets(k) = 3;
            end
            alive_matrix(1:num_targets(k),k) = 1;
        end   
    end
    
    stream = RandStream('mt19937ar','seed',seed);
    RandStream.setGlobalStream(stream);

    %% filter processing
    estimated_state = [];

    % save intermediate parameters for algorithm optimization (of CPHD filters)
    if algorithm == 3 || algorithm == 4
        fact_table = factorial(0:max_card);
        nCi_table = zeros(max_card+1);
        for n = 0:max_card
            index = 0:n;
            nCi_table(n+1,index+1) = fact_table(n+1)./(fact_table(index+1).*fact_table(n-index+1));
        end
        ps_powers = p_survive.^(0:max_card);
        one_ps_powers = (1-p_survive).^(0:max_card);
    end

    %% filter initialization
    % The J_k, w_k, m_k and P_k parameters describe the posterior PHD as a mixture of Gaussians
    % pmf_k is the probability mass function of the posterior cardinality
    
    k = 1;
    J_k(k) = 0;
    w_k{k} = [];
    m_k{k} = [];
    P_k{k} = [];
    pmf_k{k} = ones(1,max_card+1)/(max_card+1);
    
    %%
    for k = 2:k_max

        %% take measurements at sensors
        for s = 1:num_sensors
            meas = H*target_state(:,1:num_targets(k),k);
            meas = meas + sigma_e*randn(size(meas));
            ind = rand(1,num_targets(k))<pD(s);
            Z{s} = meas(:,ind);
            
            % (Poisson) clutter measurements uniform over the surveillance region
            no_clutter = poissrnd(clutter_rate);
            meas_clutter = rand(2,no_clutter)*(xrange(2)-xrange(1))-xrange(2);
            Z{s} = horzcat(Z{s},meas_clutter);
        end
        
        % for reverse sensor order processing
        if do_reverse
            p_detect = fliplr(pD);
            Z = fliplr(Z);
        else
            p_detect = pD;
        end
        
       
        %% Prediction for existing targets and new target births
        % The J_k1, w_k1, m_k1 and P_k1 parameters describe the predicted PHD as a mixture of Gaussians
        w_k1 = []; J_k1 = 0; m_k1 = []; P_k1= [];
        
        % propagate existing Gaussian components
        for j = 1:J_k(k-1)
            w_k1(j) = p_survive*w_k{k-1}(j);
            m_k1(:,j) = A*m_k{k-1}(:,j);
            P_k1(:,:,j) = Q + A*P_k{k-1}(:,:,j)*A';
        end
        
        % add new Gaussian components to account for target births
        i = J_k(k-1);
        for j = 1:J_gamma
            i = i+1;
            w_k1(i) = w_gamma(j);
            m_k1(:,i) = m_gamma(:,j);
            P_k1(:,:,i) = P_gamma;
        end
        
        J_k1 = J_k(k-1) + J_gamma;
        
        %% Prediction for cardinality distribution (convolution)
        if algorithm == 3 || algorithm == 4
            pmf_b = poisspdf(0:max_card,sum(w_gamma));
            pmf_k1 = [];

            for n = 0:max_card
                temp1 = 0;
                for j = 0:n
                    index = j:max_card;
                    temp2 = nCi_table(index+1,j+1)';
                    temp2 = temp2.*(ps_powers(j+1)*one_ps_powers(index-j+1));
                    temp2 = temp2*pmf_k{k-1}(index+1)';

                    temp1 = temp1 + pmf_b(n-j+1)*temp2;
                end
                pmf_k1(n+1) = temp1;
            end
            pmf_k1 = pmf_k1/sum(pmf_k1);
        end
        
        %% Filter Update Step
        if algorithm == 1   % multisensor iterated-corrector GM PHD
            % sensor 1
            [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = gm_phd_update(H, R, clutter_rate,...
                xrange, yrange, p_detect(1), w_k1, m_k1, P_k1, J_k1, Z{1});
            
            % pruning and merging
            [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = prune_merge(w_threshold,...
                merge_threshold, J_max, w_k{k}, m_k{k}, P_k{k});
            
            % rest of the sensors
            for s = 2:num_sensors
                [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = gm_phd_update(H, R, clutter_rate,...
                    xrange, yrange, p_detect(s), w_k{k}, m_k{k}, P_k{k}, J_k(k), Z{s});
                
                % pruning and merging
                [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = prune_merge(w_threshold,...
                    merge_threshold, J_max, w_k{k}, m_k{k}, P_k{k});
            end

        elseif algorithm == 2   % general multisensor GM PHD filter
            [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = gen_gm_ms_phd_update(H, R, clutter_rate,...
                    xrange, yrange, p_detect, w_k1, m_k1, P_k1, J_k1, W_max, PT_max, Z);
        
            % pruning and merging
            [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = prune_merge(w_threshold,...
                merge_threshold, J_max, w_k{k}, m_k{k}, P_k{k});
            
        elseif algorithm == 3 % multisensor iterated-corrector GM CPHD filter
            % sensor 1
            [pmf_k{k}, w_k{k}, m_k{k}, P_k{k}, J_k(k)] = gm_cphd_update(H, R, clutter_rate,...
                    xrange, yrange, p_detect(1), pmf_k1, w_k1, m_k1, P_k1, J_k1, Z{1});
                
            % pruning and merging
            [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = prune_merge(w_threshold,...
                merge_threshold, J_max, w_k{k}, m_k{k}, P_k{k});
            
            % rest of the sensors
            for s = 2:num_sensors
                [pmf_k{k}, w_k{k}, m_k{k}, P_k{k}, J_k(k)] = gm_cphd_update(H, R, clutter_rate,...
                    xrange, yrange, p_detect(s), pmf_k{k}, w_k{k}, m_k{k}, P_k{k}, J_k(k), Z{s});
                
                % pruning and merging
                [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = prune_merge(w_threshold,...
                    merge_threshold, J_max, w_k{k}, m_k{k}, P_k{k});
            end


        elseif algorithm == 4   % general multisensor GM CPHD filter
            [pmf_k{k}, w_k{k}, m_k{k}, P_k{k}, J_k(k)] = gen_gm_ms_cphd_update(H, R, clutter_rate,...
                    xrange, yrange, p_detect, pmf_k1, w_k1, m_k1, P_k1, J_k1, W_max, PT_max, Z);
        
            % pruning and merging
            [w_k{k}, m_k{k}, P_k{k}, J_k(k)] = prune_merge(w_threshold,...
                merge_threshold, J_max, w_k{k}, m_k{k}, P_k{k});
        end
        
        
        %% Compute the multi-target state estimates
        if  algorithm == 3 || algorithm == 4
            [~, estimated_num_targets(k)] = max(pmf_k{k});
            estimated_num_targets(k) = estimated_num_targets(k) - 1;
            
            [~, ind] = sort(w_k{k},'descend');
            estimated_num_targets(k) = min(estimated_num_targets(k),length(ind));   % to avoid array access overflow
            ind = ind(1:estimated_num_targets(k));
            
            for i = 1:estimated_num_targets(k)
                estimated_state{k}(:,i) = m_k{k}(:,ind(i));
            end
            if isempty(ind)
                estimated_state{k} = [];
            end
        else
            estimated_num_targets(k) = round(sum(w_k{k}));
            
            [~, ind] = sort(w_k{k},'descend');
            estimated_num_targets(k) = min(estimated_num_targets(k),length(ind));   % to avoid array access overflow
            ind = ind(1:estimated_num_targets(k));
            
            for i = 1:estimated_num_targets(k)
                estimated_state{k}(:,i) = m_k{k}(:,ind(i));
            end
            if isempty(ind)
                estimated_state{k} = [];
            end
        end
        
        
        %% Limit the number of components propagaed to next time step for computational efficiency
        J_k(k) = min(J_k(k),max(4*estimated_num_targets(k),4));
        w_k{k} = w_k{k}(1:J_k(k));
        m_k{k} = m_k{k}(:,1:J_k(k));
        P_k{k} = P_k{k}(:,:,1:J_k(k));
        
    end
    
    %% set the output structure parameters
    output_s.state_X = target_state;
    output_s.state_n = num_targets;
    output_s.state_X_hat = estimated_state;
    output_s.state_n_hat = estimated_num_targets;
    output_s.time = toc;
    
    %% calculate and display average OSPA error
    cutoff_c = 100;
    order_p = 1;
    error_t = zeros(1,k_max-1);
    for k = 2:k_max
        error_t(k-1) = ospa_dist(target_state(:,1:num_targets(k),k), estimated_state{k}, cutoff_c, order_p);
    end

    avg_OSPA = mean(error_t);
    disp(avg_OSPA);
    
catch exception
    % if an exception is generated, save the report in the output variable
    output_s =  getReport(exception);
end

end
